import numpy as np
import numpy.random as alea
import matplotlib.pyplot as plt
import scipy.special

###################################################################
####     Modèle d'Ehrenfest
###################################################################

# Evolution de la marche sur le cube 
def cube(a,n):
    X=np.zeros(a)
    print(X)
    for i in range(n):
        x=int(np.floor(a*alea.random(1)))
        X[x]=1-X[x]
        print(X)

# Trajectoire de la chaîne d'Ehrenfest 
# x=point de départ, a=nombre de particules, n=temps final de simulation
def ehrenfest(x,a,n):
    X=x*np.ones(n+1)
    for i in range(n):
        X[i+1]=X[i]+2*(alea.random(1)>X[i]/a)-1
    plt.plot(X)
    plt.plot([0,n],[0,0])
    plt.plot([0,n],[a/2,a/2])
    plt.plot([0,n],[a,a])
    plt.show()
    plt.title("Trajectoire de la chaîne d'Ehrenfest")

# Trajectoire de la pression 
# x=point de départ, a=nombre de particules, n=temps final de simulation
def pression(x,a,n):
    X=x*np.ones(n+1)
    for i in range(n):
        X[i+1]=X[i]+2*(alea.random(1)>X[i]/a)-1
    plt.plot(X/a)
    plt.plot([0,n],[0,0])
    plt.plot([0,n],[1/2,1/2])
    plt.plot([0,n],[1,1])
    plt.show()
    plt.title("Trajectoire de la pression")

# Trajectoire de la chaîne d'Ehrenfest jusqu'à revenir
# x=point de départ, a=nombre de particules
def excursion(x,a):
    X=[x]
    X.append(X[-1]+2*(alea.random(1)>X[-1]/a)-1)
    while X[-1]!=x:
        X.append(X[-1]+2*(alea.random(1)>X[-1]/a)-1)
    plt.plot(X)
    n=len(X)-1
    print(n)
    plt.plot([0,n],[0,0])
    plt.plot([0,n],[x,x])
    plt.plot([0,n],[a,a])
    plt.show()
    plt.title("Excursion")

#Estimation temps de retour en x
# x=point de départ, a=nombre de particules, N=taille de l'échantillon
def ehrenfestretour(x,a,N):
    T=np.ones(N)
    for i in range(N):
        X=x
        X=X+2*(alea.random(1)>X/a)-1
        c=1
        while X!=x:
            X=X+2*(alea.random(1)>X/a)-1
            c=c+1
        T[i]=c
    return(np.mean(T))


# proba inv = 1/(Espérance du temps de retour)  
# x=point de départ, a=nombre de particules, N=nbre de traj
def invarianteretour(a,N):
    Temps=np.zeros(a+1)
    Theo=np.zeros(a+1)
    for x in range(a+1):
        Temps[x]=ehrenfestretour(x,a,N)
        Theo[x]=scipy.special.binom(a, x)/(2**a)
    plt.bar(np.arange(0,a+1)-.2,1/Temps,width=.4)
    plt.bar(np.arange(0,a+1)+.2,Theo,width=.4)
    plt.title("Comparaison inverse du temps moyen de retour/loi binomiale")
    plt.show()

# retour moyen en a/2 comparé à \sqrt{\pi a/2} 
def retourmilieu(a,N):
    Temps=ehrenfestretour(a/2,a,N)
    return(Temps/np.sqrt(np.pi*a/2))

#Illustration du théorème ergodique
# x=point de départ, a=nombre de particules, n=temps final de simulation
def thergo(x,a,n):
    Eff=np.zeros(a+1)
    Theo=np.zeros(a+1)
    X=x*np.ones(n+1)
    for i in range(n):
        X[i+1]=X[i]+2*(alea.random(1)>X[i]/a)-1
    for i in range(a+1):
        Eff[i]=np.sum(X==i)
        Theo[i]=scipy.special.binom(a, i)/(2**a)
    plt.bar(np.arange(0,a+1)-.2,Eff/(n+1),width=.4)
    plt.bar(np.arange(0,a+1)+.2,Theo,width=.4)
    plt.title("Comparaison mesure empirique/loi binomiale")
    plt.show()
    
# Comportement des puissances de la matrice de transition 
def puissance(a,eps,n):
    P=np.zeros((a+1,a+1))     
    for i in range(a):
        P[i,i+1]=1-i/a
        P[i+1,i]=(i+1)/a
    P=eps*np.eye(a+1)+(1-eps)*P
    print(np.linalg.matrix_power(P,n))
    
###################################################################
####      Simulation d'une loi à support fini
###################################################################

def fini(n):
    P=[.4, .3, .2, .1]
    Q=np.cumsum(P)
    X=np.zeros(4)
    for i in range(n):
        U=alea.random(1)
        c=0
        while U>Q[c]: 
            c=c+1
        X[c]=X[c]+1
    print(X/n)    
    
###################################################################
####      Collectionneur de vignettes
###################################################################

def simuT(N):
    T=0
    for i in range(N):
        T=T+alea.geometric((N-i)/N)
    return(T)

def survieT(N,k):
    T=np.zeros(k+1)
    for i in range(k):
        T[i+1]=simuT(N)
    print(np.mean(T))
    T=np.sort(T)
    plt.step(T,np.arange(1,-1/k,-1/k),where='post')
    plt.xlabel("Nombre de battage n ")
    plt.ylabel("Proba de {T > n}")
    plt.show()
    
def survieTmultiN(k):
    N=[20,52,100,200,500,1000,10000]
    for j in range(len(N)):
        T=np.zeros(k+1)
        for i in range(k):
            T[i+1]=simuT(N[j])
        print(np.mean(T))
        T=np.sort(T)
        plt.step(T/(N[j]*np.log(N[j])),np.arange(1,-1/k,-1/k),where='post')
    plt.xlabel("Nombre de battage n ")
    plt.ylabel("Proba de {T > n}")
    plt.show()

###################################################################
####      Modèle de Wright-Fisher
###################################################################

    
# Trajectoire de la chaîne de Wright-Fisher jusqu'à toucher 0 ou a 
# x=point de départ, N=nombre d'infividus
def wrightfisher(x,N):
    X=[x]
    while X[-1]*(N-X[-1])>0:
        X.append(alea.binomial(N,X[-1]/N))
    plt.plot(X)
    n=len(X)-1
    print(n)
    plt.plot([0,n],[0,0])
    plt.plot([0,n],[N,N])
    plt.show()
    
# Temps et lieu d'absorption pour Wright-Fisher
# x=point de départ, N=nombre d'infividus
def wrightfisherabs(x,N):
    X=x
    c=0
    while X*(N-X)>0:
        X=alea.binomial(N,X/N)
        c=c+1
    return([c,X/N])
    
def wrightfishersortie(x,N,k):
    X=np.zeros(k)
    T=np.zeros(k)
    for i in range(k):
        [T[i],X[i]]=wrightfisherabs(x,N)
    return([np.mean(T),np.std(T),np.mean(X==1),np.std(X)])        
    
def wrightfishertrace(N,k):
    MT=np.zeros(N+1)
    ST=np.zeros(N+1)
    MP=np.zeros(N+1)
    SP=np.zeros(N+1)
    for i in range(N+1):
        [MT[i],ST[i],MP[i],SP[i]]=wrightfishersortie(i,N,k)
    plt.figure(figsize=(11,7))
    plt.subplot(121)
    plt.plot(np.arange(0,N+1),MT)
    plt.plot(np.arange(0,N+1),MT+1.96*ST/np.sqrt(k))
    plt.plot(np.arange(0,N+1),MT-1.96*ST/np.sqrt(k))
    plt.xlabel("Point de départ")
    plt.ylabel("Temps moyen d'absorption")
    plt.subplot(122)
    plt.plot(np.arange(0,N+1),MP)
    plt.plot(np.arange(0,N+1),MP+1.96*SP/np.sqrt(k))
    plt.plot(np.arange(0,N+1),MP-1.96*SP/np.sqrt(k))            
    plt.xlabel("Point de départ")
    plt.ylabel("Proba d'absorption en N")
    plt.show()
    
###################################################################
####      Problème de la ruine du joueur
###################################################################
        
# Trajectoire de la fortune jusqu'à toucher 0 ou N 
# x=point de départ, N=borne d'arrêt raisonnable, p=proba de victoire
def ruine(x,N,p):
    X=[x]
    while X[-1]*(N-X[-1])>0:
        X.append(X[-1]+2*(alea.random()<p)-1)
    plt.plot(X)
    n=len(X)-1
    print(n)
    plt.plot([0,n],[0,0])
    plt.plot([0,n],[N,N])
    plt.show()
    
# Temps et lieu d'absorption pour la fortune
# x=point de départ, N=borne d'arrêt raisonnable, p=proba de victoire
def ruineabs(x,N,p):
    X=x
    c=0
    while X*(N-X)>0:
        X=X+2*(alea.random()<p)-1
        c=c+1
    return([c,X/N])
    
def ruinesortie(x,N,k,p):
    X=np.zeros(k)
    T=np.zeros(k)
    for i in range(k):
        [T[i],X[i]]=ruineabs(x,N,p)
    return([np.mean(T),np.std(T),np.mean(X==1),np.std(X)])        
    
def ruinetrace(N,k,p):
    MT=np.zeros(N+1)
    ST=np.zeros(N+1)
    MP=np.zeros(N+1)
    SP=np.zeros(N+1)
    for i in range(N+1):
        [MT[i],ST[i],MP[i],SP[i]]=ruinesortie(i,N,k,p)
    plt.figure(figsize=(11,7))
    plt.subplot(121)
    plt.plot(np.arange(0,N+1),MT)
#    plt.plot(np.arange(0,N+1),MT+1.96*ST/np.sqrt(k))
#    plt.plot(np.arange(0,N+1),MT-1.96*ST/np.sqrt(k))
    plt.xlabel("Point de départ")
    plt.ylabel("Temps moyen du jeu")
    plt.subplot(122)
    plt.plot(np.arange(0,N+1),MP)
#    plt.plot(np.arange(0,N+1),MP+1.96*SP/np.sqrt(k))
#    plt.plot(np.arange(0,N+1),MP-1.96*SP/np.sqrt(k))            
    plt.xlabel("Point de départ")
    plt.ylabel("Proba de gagner N jetons")
    plt.show()
        
    
        
        
        
        
        
        